/***********************************************************
*      SAM7SExx_DevMaster - Init of external MI0283QT      *
*                                                          *
*    generated by Frank Goetze  -   www.embedded-os.de     *
*       based on CPP-library from www.watterott.net        *
***********************************************************/
#include "../inc/OS_Def_AT91SAM7xxx.h"
#include "TFTfont8x8.h"
#include "TFTfont8x12.h"


/***********************************************************
*            user-functions prototypes (MI0283QT)          *
***********************************************************/
void   TFT_Init(void);                                     // init the used SPI and the TFT
void   TFT_Reset(void);                                    // reset the TFT
void   TFT_Clear(U16 color);                               // clear screen
U08    TFT_drawPixel(U16 x, U16 y, U16 color);             // draw a pixel
U08    TFT_drawLine(U16 x0, U16 y0, U16 x1, U16 y1, U16 color);    // draw a line
U08    TFT_drawRect(U16 x0, U16 y0, U16 x1, U16 y1, U16 color);    // draw a rectangle
U08    TFT_fillRect(U16 x0, U16 y0, U16 x1, U16 y1, U16 color);    // draw a filled rectangle
U08    TFT_drawCircle(U16 x, U16 y, U16 rad, U16 color);   // draw a circle
U08    TFT_fillCircle(U16 x, U16 y, U16 rad, U16 color);   // draw a filled circle
void   TFT_gotoHome(void);                                 // set char cursor position to home (0,0)
U08    TFT_setCharCursor(U16 x, U16 y);                    // set char cursor position (0..320, 0..240)
U32    TFT_getCharCursor(void);                            // get char cursor position (0..320, 0..240)
U08    TFT_drawChar(U08 c, U08 size, U16 color, U16 bgcolor);      // draw a char at current position and inc pos
U08    TFT_drawString(U08 *s, U08 size, U16 color, U16 bgcolor);   // draw a string and inc pos
U08    TFT_changeFont(U08 font);                           // change font (8x8 or 8x12)

/***********************************************************
*                    definitions                           *
***********************************************************/
#define CPU_CRISTAL_CLK       18432000UL                   // processor crystal-clock in Hz

#define AT91S_SPI_NPCS2       AT91S_PB30                   // /MCS2 PIO-pin
#define AT91S_SPI_NPCS3       AT91S_PB22                   // /MCS3 PIO-pin
#define AT91S_SPI_MISO        AT91S_PIO_PA12               // MISO  PIO-pin
#define AT91S_SPI_MOSI        AT91S_PIO_PA13               // MOSI  PIO-pin
#define AT91S_SPI_SPCK        AT91S_PIO_PA14               // SCK   PIO-pin
#define AT91S_TFT_BL          AT91S_PB29                   // LCD-backlight PIO-pin
#define AT91S_TFT_RES         AT91S_PB31                   // LCD-Reset PIO-pin
#define AT91S_PEN_IRQ         AT91S_PIO_PA30               // ADS-Pen IRQ-pin 

#define TFT_MAX_X             320
#define TFT_MAX_Y             240
#define RGB(r,g,b)            (((r&0xF8)<<8)|((g&0xFC)<<3)|((b&0xF8)>>3)) // 5 red | 6 green | 5 blue
#define TFT_ID                (0)
#define TFT_DATA              ((0x72)|(TFT_ID<<2))
#define TFT_REGISTER          ((0x70)|(TFT_ID<<2))

#define TFTport_CS_ON         AT91S_PIOB_CODR = AT91S_SPI_NPCS2;
#define TFTport_CS_OFF        AT91S_PIOB_SODR = AT91S_SPI_NPCS2;
#define ADSport_CS_ON         AT91S_PIOB_CODR = AT91S_SPI_NPCS3;
#define ADSport_CS_OFF        AT91S_PIOB_SODR = AT91S_SPI_NPCS3;
#define TFTport_BL_ON         AT91S_PIOB_SODR = AT91S_TFT_BL;
#define TFTport_BL_OFF        AT91S_PIOB_CODR = AT91S_TFT_BL;
#define TFTport_drawStart     TFTport_CS_ON; TFTport_Byte(TFT_REGISTER); TFTport_Byte(0x22); TFTport_CS_OFF; TFTport_CS_ON; TFTport_Byte(TFT_DATA)
#define TFTport_drawEnd       TFTport_CS_OFF


#define TFT_NO_ERR            0
#define TFT_PARAM_ERR         1

/***********************************************************
*                       globals                            *
***********************************************************/
OS_CONST U08 OS_ROMDATA TFT_InitParams[] = { 0xEA, 0xEB, 0xEC, 0xED, 0xE8, 0xE9, 0xF1, 0xF2, 0x27, 0x1B, 0x1A, 0x24, 0x25, 0x23, 0x18, 0x19, 0x01, 0x1F, 0x1F, 0x1F, 0x1F, 0x17, 0x36, 0x28, 0x28, 0x16, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09,     
                                             0x00, 0x20, 0x0C, 0xC4, 0x40, 0x38, 0x01, 0x10, 0xA3, 0x1B, 0x01, 0x2F, 0x57, 0x8D, 0x36, 0x01, 0x00, 0x88, 0x80, 0x90, 0xD0, 0x05, 0x00, 0x38, 0x3C, 0xA8, 0x00, 0x00, 0x01, 0x3F, 0x00, 0x00, 0x00, 0xEF   };

U32  TFTxy;                                                // char cursor-position
U08 *TFTfont;                                              // selected font
U08  TFTfont_height;                                       // height of selected font
U08  TFTfont_width;                                        // width of selected font

/***********************************************************
*          get MCK based on given cristal-CLK              *
***********************************************************/
U32 gl_mck;                                                // global value holder
void OS_getMCK(void)
{
    U32 s_clksrc;
    
    s_clksrc = CPU_CRISTAL_CLK;                                        // set start-clk (cristal-clk) - must be a "define"
    if((AT91S_PMC_MCKR & AT91S_PMC_PLLCLK) == AT91S_PMC_PLLCLK) {      // if PLL is enabled ..
        if((AT91S_CKGR_PLLR & AT91S_CKGR_DIV) > 0) {                   // .. if PLL-divisor set ..
            s_clksrc /= (AT91S_CKGR_PLLR & AT91S_CKGR_DIV);            // div clk with PLL-div
        }
        s_clksrc *= (((AT91S_CKGR_PLLR & AT91S_CKGR_MUL) >> 16) + 1);  // mul clk with PLL-multiplier
    }
    s_clksrc /= (0x0001UL << ((AT91S_PMC_MCKR & AT91C_PMC_PRES_CLK_M) >> 2));  // div clk with MCK-div
    gl_mck   = s_clksrc;                                               // now "gl_mclk" holds the MCK in Hertz
}

/***********************************************************
*       set max possible SPI baudrate and return this      *
***********************************************************/
U32 OS_setSPI_BR(U32 maxclk)
{
    U08 i, m;

    OS_getMCK();
    AT91S_SPI_MR &= ~AT91S_SPI_MR_FDIV;                    // clear FDIV at calc-startup
    for(i=1, m=1; m<32; i=1, m=32) {                       // search lowest clk-div for SPI-speed (highest possible SPI-speed)
        while(((gl_mck / (i * m)) > maxclk) && (i > 0)) i++;
        if(i > 0) break;
    }
    if(m > 1) AT91S_SPI_MR |= AT91S_SPI_MR_FDIV;           // if hole SPI-div greater 255 set FDIV for enable pre-div(32)
    AT91S_SPI_CSR->s[2] = (AT91S_SPI_CSR->s[2] & ~AT91S_SPI_CSR_SCBR_M) | (i << 8);
    return(gl_mck / (i * m));                              // set SPI-div and return the used SPI-speed in Hz
}

/***********************************************************
*                     simple delay                         *
***********************************************************/
void TFTport_Delay(U32 d)
{
    d *= 10;
    while(d--) { _NOP(); _NOP(); _NOP(); }
}

/***********************************************************
*                  write a U08 to display                  *
***********************************************************/
void TFTport_Byte(U08 w)
{
    AT91S_SPI_TDR = w;
    while(!(AT91S_SPI_SR & AT91S_SPI_SR_TDRE));            // wait up to tx-empty
}

/***********************************************************
*                  write a U16 to display                  *
***********************************************************/
void TFTport_Word(U16 w)
{
    AT91S_SPI_TDR = (w >> 8);
    while(!(AT91S_SPI_SR & AT91S_SPI_SR_TDRE));            // wait up to tx-empty
    AT91S_SPI_TDR = w;
    while(!(AT91S_SPI_SR & AT91S_SPI_SR_TDRE));            // wait up to tx-empty
}

/***********************************************************
*                      write a CMD                         *
***********************************************************/
void TFTport_Cmd(U08 reg, U08 param)
{
    TFTport_CS_ON;
    TFTport_Byte(TFT_REGISTER);
    TFTport_Byte(reg);
    TFTport_CS_OFF;
    TFTport_CS_ON;
    TFTport_Byte(TFT_DATA);
    TFTport_Byte(param);
    TFTport_CS_OFF;
}

/***********************************************************
*                  set area for update                     *
***********************************************************/
void TFTport_setArea(U16 x0, U16 y0, U16 x1, U16 y1)
{
    TFTport_Cmd(0x02, (x0>>8)); TFTport_Cmd(0x03, x0); //set x0
    TFTport_Cmd(0x04, (x1>>8)); TFTport_Cmd(0x05, x1); //set x1
    TFTport_Cmd(0x06, (y0>>8)); TFTport_Cmd(0x07, y0); //set y0
    TFTport_Cmd(0x08, (y1>>8)); TFTport_Cmd(0x09, y1); //set y1
}

/***********************************************************
*                     clear screen                         *
***********************************************************/
void TFT_Clear(U16 color)
{
    U32 i;
    
    TFTport_setArea(0, 0, TFT_MAX_X - 1, TFT_MAX_Y - 1);
    TFTport_drawStart;
    for(i=0; i < (TFT_MAX_X*TFT_MAX_Y); i++) TFTport_Word(color);
    TFTport_drawEnd;
    TFTxy = 0;
}

/***********************************************************
*                     draw a pixel                         *
***********************************************************/
U08 TFT_drawPixel(U16 x, U16 y, U16 color)
{
    if ((x >= TFT_MAX_X) || (y >= TFT_MAX_Y))
        return TFT_PARAM_ERR;
    TFTport_setArea(x, y, x, y);
    TFTport_drawStart;
    TFTport_Word(color);
    TFTport_drawEnd;
    return TFT_NO_ERR;
}

/***********************************************************
*                draw a filled rectangle                   *
***********************************************************/
U08 TFT_fillRect(U16 x0, U16 y0, U16 x1, U16 y1, U16 color)
{
    U32    tmp;
    
    if ((x0 >= TFT_MAX_X) || (y0 >= TFT_MAX_Y) || (x1 >= TFT_MAX_X) || (y1 >= TFT_MAX_Y))
        return TFT_PARAM_ERR;
    if (x0 > x1) {
        tmp = x0; x0 = x1, x1 = tmp;
    }
    if (y0 > y1) {
        tmp = y0; y0 = y1, y1 = tmp;
    }
    TFTport_setArea(x0, y0, x1, y1);
    TFTport_drawStart;
    for (tmp = ((1 + (x1 - x0)) * (1 + (y1 - y0))); tmp != 0; tmp--)
        TFTport_Word(color);
    TFTport_drawEnd;
    return TFT_NO_ERR;
}

/***********************************************************
*                   draw a rectangle                       *
***********************************************************/
U08 TFT_drawRect(U16 x0, U16 y0, U16 x1, U16 y1, U16 color)
{
    if ((x0 >= TFT_MAX_X) || (y0 >= TFT_MAX_Y) || (x1 >= TFT_MAX_X) || (y1 >= TFT_MAX_Y))
        return TFT_PARAM_ERR;
    TFT_fillRect(x0, y0, x0, y1, color);
    TFT_fillRect(x0, y1, x1, y1, color);
    TFT_fillRect(x1, y0, x1, y1, color);
    TFT_fillRect(x0, y0, x1, y0, color);
    return TFT_NO_ERR;
}

/***********************************************************
*                     draw a line                          *
***********************************************************/
U08 TFT_drawLine(U16 x0, U16 y0, U16 x1, U16 y1, U16 color)
{
    S16   dx, dy, dx2, dy2, err, sx, sy;

    if ((x0 >= TFT_MAX_X) || (y0 >= TFT_MAX_Y) || (x1 >= TFT_MAX_X) || (y1 >= TFT_MAX_Y))
        return TFT_PARAM_ERR;
    if ((x0 == x1) || (y0 == y1))
        return (TFT_fillRect(x0, y0, x1, y1, color));
    dx = x1 - x0; dy = y1 - y0;
    if(dx < 0) { dx = -dx; sx = -1; } else { sx = +1; }
    if(dy < 0) { dy = -dy; sy = -1; } else { sy = +1; }
    dx2 = dx << 1; dy2 = dy << 1;
    TFT_drawPixel(x0, y0, color);
    if(dx > dy) {
        err = dy2 - dx;
        while(x0 != x1) {
            if(err >= 0) { y0  += sy; err -= dx2; }
            x0  += sx; err += dy2;
            TFT_drawPixel(x0, y0, color);
        }
    } else {
        err = dx2 - dy;
        while(y0 != y1) {
            if(err >= 0) {  x0  += sx; err -= dy2; }
            y0  += sy; err += dx2;
            TFT_drawPixel(x0, y0, color);
        }
    }
    return TFT_NO_ERR;
}

/***********************************************************
*                    draw a circle                         *
***********************************************************/
U08 TFT_drawCircle(U16 x, U16 y, U16 rad, U16 color)
{
    S16 xd, yd, err;
    
    if ((x >= TFT_MAX_X) || (y >= TFT_MAX_Y))
        return TFT_PARAM_ERR;
    if (((x + rad) >= TFT_MAX_X) || ((x - rad) >= TFT_MAX_X))
        return TFT_PARAM_ERR;
    if (((y + rad) >= TFT_MAX_Y) || ((y - rad) >= TFT_MAX_Y))
        return TFT_PARAM_ERR;
    xd = rad; yd = 0; err = -rad;
    while (xd >= yd) {
        TFT_drawPixel(x + xd, y + yd, color);
        TFT_drawPixel(x - xd, y + yd, color);
        TFT_drawPixel(x + xd, y - yd, color);
        TFT_drawPixel(x - xd, y - yd, color);
        TFT_drawPixel(x + yd, y + xd, color);
        TFT_drawPixel(x - yd, y + xd, color);
        TFT_drawPixel(x + yd, y - xd, color);
        TFT_drawPixel(x - yd, y - xd, color);
        err += yd; yd++; err += yd;
        if (err >= 0) { xd--; err -= xd; err -= xd; }
    }
    return TFT_NO_ERR;
}

/***********************************************************
*                 draw a filled circle                     *
***********************************************************/
U08 TFT_fillCircle(U16 x, U16 y, U16 rad, U16 color)
{
    S16 xd, yd, err;
    
    if ((x >= TFT_MAX_X) || (y >= TFT_MAX_Y))
        return TFT_PARAM_ERR;
    if (((x + rad) >= TFT_MAX_X) || ((x - rad) >= TFT_MAX_X))
        return TFT_PARAM_ERR;
    if (((y + rad) >= TFT_MAX_Y) || ((y - rad) >= TFT_MAX_Y))
        return TFT_PARAM_ERR;
    xd = rad; yd = 0; err = -rad;
    while (xd >= yd) {
        TFT_fillRect(x + xd, y + yd, x - xd, y + yd, color);
        TFT_fillRect(x + xd, y - yd, x - xd, y - yd, color);
        TFT_fillRect(x + yd, y + xd, x - yd, y + xd, color);
        TFT_fillRect(x + yd, y - xd, x - yd, y - xd, color);
        err += yd; yd++; err += yd;
        if (err >= 0) { xd--; err -= xd; err -= xd; }
    }
    return TFT_NO_ERR;
}

/***********************************************************
*            set char cursor position to home              *
***********************************************************/
void TFT_gotoHome(void)
{
    TFTxy = 0;
}

/***********************************************************
*         set char cursor position (0..320, 0..240)        *
***********************************************************/
U08 TFT_setCharCursor(U16 x, U16 y)
{
    if ((x >= TFT_MAX_X) || (y >= TFT_MAX_Y))
        return TFT_PARAM_ERR;
    TFTxy = ((U32)(x) << 16) | y;
    return TFT_NO_ERR;
}

/***********************************************************
*         get char cursor position (0..320, 0..240)        *
***********************************************************/
U32 TFT_getCharCursor(void)
{
    return(TFTxy);
}

/***********************************************************
*       draw a char at current position and inc pos        *
*  --> no transparent background possible (no read bit) !  *
***********************************************************/
U08 TFT_drawChar(U08 c, U08 size, U16 color, U16 bgcolor)
{
    U16    i, x, y, xs, ys, bm;
    U08    mask;
    
    x = (U16)(TFTxy >> 16);
    y = (U16)(TFTxy & 0xFFFF);
    if (size == 0) size = 1;
    if ((x + (TFTfont_width * size)) > TFT_MAX_X) {
        x  = 0; y += (TFTfont_height * size);
    }
    if ((y + (TFTfont_height * size)) > TFT_MAX_X) {
        x = 0; y = 0;
    }
    TFTport_setArea(x, y, x + (TFTfont_width * size) - 1, y + (TFTfont_height * size) - 1);
    TFTport_drawStart;
    for (i = 0; i < TFTfont_height; i++) {
        bm = (c >= 0x20)? TFTfont[TFTfont_height * (c - 0x20) + i] : 0;
        for (ys = size; ys != 0; ys--) {
            for (mask = 0x80; mask != 0; mask >>= 1) {
                for (xs = size; xs != 0; xs--) TFTport_Word((bm & mask)? color : bgcolor);
            }
        }
    }
    TFTport_drawEnd;
    TFTxy = ((U32)(x + (TFTfont_width * size)) << 16) | y;
    return TFT_NO_ERR;
}

/***********************************************************
*              draw a string and inc pos                   *
***********************************************************/
U08 TFT_drawString(U08 *s, U08 size, U16 color, U16 bgcolor)
{
    U16 i, r;
    
    i = 0; r = 0;
    while((s[i] != 0) && (r == 0))
        r = TFT_drawChar(s[i++], size, color, bgcolor);
    return TFT_NO_ERR;
}

/***********************************************************
*              change font (8x8 or 8x12)                   *
***********************************************************/
U08  TFT_changeFont(U08 font)
{
    switch (font) {
    case 0:  
        TFTfont        = (U08 *)TFTfont8x8;
        TFTfont_height = TFTfont8x8_height;
        TFTfont_width  = TFTfont8x8_width;
        break;
    case 1:   
        TFTfont        = (U08 *)TFTfont8x12;
        TFTfont_height = TFTfont8x12_height;
        TFTfont_width  = TFTfont8x12_width;
        break;
    default:  
        return TFT_PARAM_ERR;
        //lint -unreachable
//      break;
    }
    return TFT_NO_ERR;
}

/***********************************************************
*                     reset the LCD                        *
***********************************************************/
void TFT_Reset(void)
{
    U08 i, r;

    AT91S_PIOB_CODR = AT91S_TFT_RES;
    TFTport_Delay(100);
    AT91S_PIOB_SODR = AT91S_TFT_RES;
    TFTport_Delay(500);
    r = 0;
    for (i = 0; i < (sizeof(TFT_InitParams)/2); i++) {
        if (r == TFT_InitParams[i]) TFTport_Delay(100);
        else r = TFT_InitParams[i];
        TFTport_Cmd(r, TFT_InitParams[(sizeof(TFT_InitParams)/2) + i]);
    }
    TFTport_Delay(500);
    TFTfont        = (U08 *)TFTfont8x8;
    TFTfont_height = TFTfont8x8_height;
    TFTfont_width  = TFTfont8x8_width;
    TFTxy          = 0;
    TFTport_BL_ON;
}

/***********************************************************
*                 SPI port for LCD init                    *
***********************************************************/
void TFTport_InitSPI(void)
{
    AT91S_PIOB_PER       = AT91S_SPI_NPCS2 | AT91S_SPI_NPCS3 | AT91S_TFT_BL | AT91S_TFT_RES;    // enable GPIO of /CS & BackLight & Reset
    AT91S_PIOB_SODR      = AT91S_SPI_NPCS2 | AT91S_SPI_NPCS3 | AT91S_TFT_RES;                   // set high
    AT91S_PIOB_CODR      = AT91S_TFT_BL;                                                        // set low
    AT91S_PIOB_OER       = AT91S_SPI_NPCS2 | AT91S_SPI_NPCS3 | AT91S_TFT_BL | AT91S_TFT_RES;    // output enable
    AT91S_PMC_PCER       = (1 << AT91S_ID_PIOB);                                                // enable clock on PIO-B

    AT91S_PIO_PDR        = AT91S_SPI_MISO | AT91S_SPI_MOSI | AT91S_SPI_SPCK;  // disable from PIO-mode
    AT91S_PIO_ASR        = AT91S_SPI_MISO | AT91S_SPI_MOSI | AT91S_SPI_SPCK;  // enable bit-functions
    AT91S_PMC_PCER       = (1 << AT91S_ID_SPI);                               // enable clock on SPI-interface
    AT91S_SPI_CR         = AT91S_SPI_CR_DIS;                                  // disable SPI
    AT91S_SPI_CR         = AT91S_SPI_CR_EN | AT91S_SPI_CR_SWRST;              // enable & reset SPI
    AT91S_SPI_MR         = AT91S_SPI_MR_CS2 | AT91S_SPI_MR_MODFDIS | AT91S_SPI_MR_MSTR;         // CS2 + fault-off + (fix) + master
    AT91S_SPI_CSR->s[2]  = AT91S_SPI_CSR_SCBR_L | AT91S_SPI_CSR_BITS_8 | AT91S_SPI_CSR_CPOL;    // low-speed +  8bit + clk-pol (mode 3)
    AT91S_SPI_CR         = AT91S_SPI_CR_EN;                                   // enable SPI
    OS_setSPI_BR(50000000);                                                   // set max 50MHz
}

/***********************************************************
*                     TFT init (MI0283QT)                  *
***********************************************************/
void TFT_Init(void)
{
    TFTport_InitSPI();
    TFT_Reset();
}

/************************* END ***************************/



/***************************************************************************
****************************************************************************
* TEST MAIN                                                                *
****************************************************************************
***************************************************************************/
void main(void)
{
    TFT_Init();
    TFT_Clear(0xFFFF);

    TFT_drawRect(38, 38, 281, 230, 0x0000);
    TFT_fillRect(40, 40, 279, 199, 0xA000);
    TFT_fillCircle(140, 120, 30, 0x0F00);
    TFT_drawCircle(180, 110, 40, 0x3313);
    TFT_drawLine(10, 80, 309, 160, 0x0300);
    TFT_drawLine(50, 160, 270, 80, 0x00F0);
    
    TFT_gotoHome();
    TFT_setCharCursor(10, 5);
    TFT_drawString("String 8x8 Test!", 1, 0x000F, 0xFFFF);
    TFT_changeFont(1);
    TFT_setCharCursor(10, 20);
    TFT_drawString("String 8x12 Test!", 1, 0x000F, 0xFFFF);
    TFT_changeFont(0);
    TFT_setCharCursor(53, 45);
    TFT_drawString("String 8x8(2)", 2, 0x9FAF, 0xA000);
    TFT_changeFont(1);
    TFT_setCharCursor(47, 165);
    TFT_drawString("String 8x12(2)", 2, 0x9FAF, 0xA000);
    TFT_changeFont(0);
    TFT_setCharCursor(50, 210);
    TFT_drawString("embedded-os.de", 2, 0x0000, 0xFFFF);
    
    while(1);
}


